# **********************************************************************
# Copyright 2019 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************

import json
from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.lcp_common.netopa.client import get_netopa_cli_output
from cli.commands.common.arguments import IntegerArgument
from cli.commands.common.arguments import IpAddressOptionalPortArgument

MAX_FRAGMENT_MIN_VAL = 1000
MAX_FRAGMENT_MAX_VAL = 65535
COLLECTOR_INTERVAL_MIN_VAL = 1
COLLECTOR_INTERVAL_MAX_VAL = 1800

def load_response_output(response):
    try:
        if response is None:
            return None
        output = json.loads(response)
    except (TypeError, ValueError) as err:
        lg.error("load response error: %s", err,
                 error_code=ec.INTERNAL_ERROR)
        return None
    return output


class MaxFragmentArgument(IntegerArgument):
    _MIN_VALUE = MAX_FRAGMENT_MIN_VAL
    _MAX_VALUE = MAX_FRAGMENT_MAX_VAL


class CollectorIntervalArgument(IntegerArgument):
    _MIN_VALUE = COLLECTOR_INTERVAL_MIN_VAL
    _MAX_VALUE = COLLECTOR_INTERVAL_MAX_VAL


class GetLogLevel(CliCommandProvider):
    def __call__(self, cmd_list):
        response = get_netopa_cli_output("get/logLevel")
        if response is None:
            return "MSG_NETOPA_LOG_LEVEL_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class SetLogLevel(CliCommandProvider):
    def __call__(self, cmd_list):
        log_level = cmd_list[4]
        response = get_netopa_cli_output("set/logLevel",
                                         [log_level])
        if response is None:
            return "MSG_NETOPA_LOG_LEVEL_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class GetCollectorAddrConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        # support multiple collector get
        response = get_netopa_cli_output("get/collectorAddrConfig")
        output = load_response_output(response)
        if response is None:
            return "MSG_NETOPA_COLLECTOR_ADDR_CONFIG_FAILED"
        return "OK\n\n%s" % json.dumps(output)


class SetCollectorAddrConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        # support multiple collector set
        ip_port_list = cmd_list[4].split(",")
        cli_arg = []
        [cli_arg.extend(ip_port.split(":")) for ip_port in ip_port_list
                                            if IpAddressOptionalPortArgument()
                                               .verify(ip_port, {})]
        if cli_arg:
            response = get_netopa_cli_output("set/collectorAddrConfig",
                                         cli_arg)
            if response is None:
                return "MSG_NETOPA_COLLECTOR_ADDR_CONFIG_FAILED"
            return "OK\n\n%s" % json.dumps(response)
        else:
            return "MSG_NETOPA_COLLECTOR_ADDR_CONFIG_FAILED"


class ClearCollectorAddrConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        response = get_netopa_cli_output("clear/collectorAddrConfig")
        if response is None:
            return "MSG_NETOPA_COLLECTOR_ADDR_CONFIG_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class GetNetopaFsmStatus(CliCommandProvider):
    def __call__(self, cmd_list):
        response = get_netopa_cli_output("get/fsmStatus")
        if response is None:
            return "MSG_NETOPA_FSM_STATUS_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class GetMaxFragmentLen(CliCommandProvider):
    def __call__(self, cmd_list):
        response = get_netopa_cli_output("get/maxFragmentLen")
        if response is None:
            return "MSG_NETOPA_MAX_FRAGMENT_SIZE_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class SetMaxFragmentLen(CliCommandProvider):
    def __call__(self, cmd_list):
        max_fragment = cmd_list[4]
        response = get_netopa_cli_output("set/maxFragmentLen",
                                         [max_fragment])
        if response is None:
            return "MSG_NETOPA_MAX_FRAGMENT_SIZE_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class GetCollectorIntervalConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        response = get_netopa_cli_output("get/collectorIntervalConfig")
        if response is None:
            return "MSG_NETOPA_COLLECTOR_INTERVAL_CONFIG_FAILED"
        return "OK\n\n%s" % json.dumps(response)


class SetCollectorIntervalConfig(CliCommandProvider):
    def __call__(self, cmd_list):
        interval = cmd_list[4]
        response = get_netopa_cli_output("set/collectorIntervalConfig",
                                         [interval])
        if response is None:
            return "MSG_NETOPA_COLLECTOR_INTERVAL_CONFIG_FAILED"
        return "OK\n\n%s" % json.dumps(response)
