# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import datetime
import gzip
import nsx_logging
import os
import re
import sys
from stat import ST_MTIME
from cli.utils.error_code import ErrorCode as ec
from cli.utils.location_utils import FilePaths
from cli.utils.render_utils import generate_datetime_stamp_str


lg = nsx_logging.getLogger("cli.utils.logging_utils")


CLEAN_EXTENSIONS = [".log", ".tgz"]

_CLI_LOG_EXTENSION = "nsxcli.log"
_LOG_DESCRIPTOR_KEYWORD = "cli.descriptors.cli_command_service"
_LOG_CMD_KEYWORD = "CMD:"


_LOG_DATE_PATTERN = "(\d+-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d+Z)"
_LOG_PID_PATTERN = "(\d+)"
_LOG_LEVEL_PATTERN = "(INFO|DEBUG|WARNING|ERROR|CRITICAL|NOTSET)"
_LOG_MODULE_PATTERN = "([^\s]+)"
_LOG_MESSAGE_PATTERN = "(.*)"
_LOG_UID_PATTERN = "{(.+)}"
_LOG_COMPRESS_EXTENSION = ".gz"


_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "fatal": ("Designates very severe events that may cause application to"
              " abort"),
    "error": ("Designates error events that may allow application to continue"
              " running"),
    "warn": "Designates potentially harmful situations",
    "info": "Designates coarse-grained informational messages",
    "debug": ("Designates fine-grained informational messages for debugging"
              " purposes"),
    "trace": "Designates finer-grained informational events than debug"
}


class LoggingLevelArgumentDescription(object):
    """Value description for LoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _LOGGING_LEVEL_DESC.get(value, "NSX logging level argument")

_EDGE_ROUTING_SERVICE_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "debug": ("Designates fine-grained informational messages for debugging"
                  " purposes"),
    "error": ("Designates error events that may allow application to continue"
               " running"),
    "informational": "Designates coarse-grained informational messages",
    "notification": "Normal but significant conditions",
    "warning": "Designates potentially harmful situations",
}


class RoutingServiceLoggingLevelArgumentDescription(object):
    """Value description for EdgeRoutingServiceLoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_SERVICE_LOGGING_LEVEL_DESC.get(
            value, "NSX Routing Service logging level argument")


_EDGE_ROUTING_PLATFORM_LOGGING_LEVEL_DESC = {
    "off": "The highest possible rank and is intended to turn off logging",
    "fatal": ("Designates very severe events that may cause application to"
              " abort"),
    "error": ("Designates error events that may allow application to continue"
              " running"),
    "warn": "Designates potentially harmful situations",
    "info": "Designates coarse-grained informational messages",
    "debug": ("Designates fine-grained informational messages for debugging"
              " purposes"),
}


class RoutingPlatformLoggingLevelArgumentDescription(object):
    """Value description for RoutingPlatformLoggingLevelArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_PLATFORM_LOGGING_LEVEL_DESC.get(
            value, "NSX Routing Platform logging level argument")


_EDGE_ROUTING_SERVICE_LOGGING_DESTINATION_DESC = {
    "file": "Routing service log",
    "syslog": "Syslog"
}


class RoutingServiceLoggingDestinationArgumentDescription(object):
    """Value description for EdgeRoutingServiceLoggingDestinationArgument."""
    @staticmethod
    def value_description(value, req_dict, factory):
        return _EDGE_ROUTING_SERVICE_LOGGING_DESTINATION_DESC.get(
            value, "NSX Routing Service logging destination argument")


def clear_log_directory(log_file_dir, stale_ts):
    """Delete log files that are older than specified timestamp.

    Args:
        log_file_dir: Directory where log files can be found.
        stale_ts: Delete log files last modified prior to this timestamp.
    """
    try:
        log_files = os.listdir(log_file_dir)
        for log_file in log_files:
            _, ext = os.path.splitext(log_file)
            if ext not in CLEAN_EXTENSIONS:
                continue
            full_path = os.path.join(log_file_dir, log_file)
            st = os.stat(full_path)
            last_modified_ts = st[ST_MTIME]
            if last_modified_ts < stale_ts:
                try:
                    os.unlink(full_path)
                except Exception:
                    lg.exception("Error deleting stale log file %s:", full_path,
                                 error_code=ec.INTERNAL_ERROR)
    except Exception:
        lg.exception("Error listing log directory files:",
                     error_code=ec.INTERNAL_ERROR)


def build_structured_data_map(security=False, audit=False, errorCode=0,
                              suppress_syslog_audit_entry=False, username=None):
    """Build structured data for logging records.
    """
    key = "tags"
    sd_map = {key: {}}
    if security:
        sd_map[key]["security"] = "true"
    if audit:
        sd_map[key]["audit"] = "true"
    if errorCode:
        sd_map[key]["errorCode"] = errorCode
    if suppress_syslog_audit_entry:
        sd_map[key]["suppress_syslog_audit_entry"] = True
    if username is not None:
        sd_map[key]["username"] = username
    return sd_map


def get_log_file_path():
    """Get the list of log files path location
    Returns:
        list of log files path
    """
    log_file_path = []
    for dirpath, dirs, files in os.walk(FilePaths.NSX_LOG_DIR):
        for filename in files:
            file_path = os.path.join(dirpath, filename)
            if filename.startswith(_CLI_LOG_EXTENSION):
                log_file_path.append(file_path)
    return log_file_path


def get_log_contents(file_path):
    """Get the list of lines of file content
    Returns:
        list of string
    """
    lines = []
    if file_path.endswith(_LOG_COMPRESS_EXTENSION):
        with gzip.open(file_path) as file_handler:
            lines = file_handler.readlines()
    else:
        with open(file_path) as file_handler:
            lines = file_handler.readlines()
    return lines


def get_command_details(file_path, userid):
    """Get the list of commands dictionary for current log file
    Args:
        file_path: string value
        userid: userid of user
    Returns:
        list of commands dictionary for current log file
    """
    cmd_details = []
    lines = get_log_contents(file_path)
    for line in lines:
        line = line.decode("utf-8") if (sys.version_info >= (3,)
            and hasattr(line, "decode")) else line
        item = re.search(('%s %s %s %s %s %s' % (_LOG_DATE_PATTERN,
            _LOG_PID_PATTERN, _LOG_MODULE_PATTERN, _LOG_LEVEL_PATTERN,
            _LOG_UID_PATTERN, _LOG_MESSAGE_PATTERN)), line)
        if item:
            item = item.groups()
            # capture command information in cmd_details list
            if (item[2].startswith(_LOG_DESCRIPTOR_KEYWORD)
                and item[5].startswith(_LOG_CMD_KEYWORD)):
                if item[4] == str(userid):
                    cmd_time = datetime.datetime.\
                        strptime(item[0], '%Y-%m-%dT%H:%M:%S.%fZ')
                    if sys.version_info >= (3,):
                        cmd_time = cmd_time.\
                            replace(tzinfo=datetime.timezone.utc).timestamp()
                    else:
                        cmd_time = (cmd_time - datetime.datetime(1970, 1, 1)). \
                            total_seconds()
                    cmd_time = generate_datetime_stamp_str(cmd_time)
                    cmd_dict = {
                        "date": cmd_time,
                        "command": item[5].split(_LOG_CMD_KEYWORD)[1].strip()
                        }
                    cmd_details.append(cmd_dict)
    return cmd_details


def get_commands_history(userid, size=0):
    """
    Get the list of commands dictionary.
    Dictionary contains information about command i.e. date, command
    Args:
        size: int value
            if 0 then return all commands list for given userid
        userid: userid of user
    Returns:
        list of commands dictionary
    """
    result = []
    log_file_path = sorted(get_log_file_path())
    for log_file_entry in log_file_path:
        result = get_command_details(log_file_entry, userid) + result
        user_cmds = []
        for item in result:
            # if user enters same command multiple times
            # in a row then display only the latest command.
            if user_cmds and user_cmds[-1]["command"] == item["command"]:
                user_cmds[-1] = item
            else:
                user_cmds.append(item)
        result = user_cmds
        if 0 < size < len(result):
            result = result[- size:]
            break
    return result
