# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import re
import sha.config
import sys
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache
from subprocess import CalledProcessError

from nsx_utils.node_utils import get_node_type
from nsx_utils.constants import NODE_TYPE_EDGE
from nsx_utils.constants import NODE_TYPE_HOST_ESX
from nsx_utils.constants import NODE_TYPE_HOST_PUBLIC_CLOUD
from nsx_utils.constants import NODE_TYPE_HOST_RHELKVM
from nsx_utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from nsx_utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from nsx_utils.constants import NODE_TYPE_MANAGER
from nsx_utils.constants import NODE_TYPE_POLICY
from nsx_utils.constants import NODE_TYPE_CONTROLLER
from nsx_utils.constants import NODE_TYPE_GLOBAL_MANAGER

# TODO: should  add it in the nsx_utils
NODE_TYPE_UA = "nsx-ua"

_NODE_TYPE = get_node_type()

if NODE_TYPE_MANAGER in _NODE_TYPE and NODE_TYPE_POLICY in _NODE_TYPE \
        and NODE_TYPE_CONTROLLER in _NODE_TYPE:
    _NODE_TYPE = NODE_TYPE_UA

_SUPPORTED_NODE_TYPES = [
    NODE_TYPE_HOST_ESX,
    NODE_TYPE_HOST_UBUNTUKVM,
    NODE_TYPE_HOST_RHELKVM,
    NODE_TYPE_EDGE,
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY,
    NODE_TYPE_UA,
    NODE_TYPE_GLOBAL_MANAGER
]

VAR_LOG_DIR = '/var/log/'

AUTH_LOG = 'auth.log'
NSX_AUDIT_LOG = 'nsx-audit.log'
NSX_AUDIT_WRITE_LOG = 'nsx-audit-write.log'
NSX_SYSLOG_LOG = 'nsx-syslog.log'
RSYSLOG_LOG = 'rsyslog.log'
SYSLOG = 'syslog'
SYSLOG_LOG = 'syslog.log'
VMWARE_NSX_SYSLOG = 'vmware/nsx-syslog'

NODE_TYPE_LOGS = {
    NODE_TYPE_HOST_ESX: [AUTH_LOG, NSX_SYSLOG_LOG],
    NODE_TYPE_HOST_UBUNTUKVM: [AUTH_LOG, VMWARE_NSX_SYSLOG],
    NODE_TYPE_HOST_RHELKVM: [VMWARE_NSX_SYSLOG],
    NODE_TYPE_EDGE: [AUTH_LOG, RSYSLOG_LOG, SYSLOG],
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY: [AUTH_LOG, RSYSLOG_LOG, SYSLOG],
    NODE_TYPE_UA: [AUTH_LOG, RSYSLOG_LOG, SYSLOG, NSX_AUDIT_LOG,
                   NSX_AUDIT_WRITE_LOG],
    NODE_TYPE_GLOBAL_MANAGER: [AUTH_LOG, RSYSLOG_LOG, SYSLOG, NSX_AUDIT_LOG,
                   NSX_AUDIT_WRITE_LOG],
}

NODE_TYPE_LOG_DIR_STATS = {
    NODE_TYPE_HOST_UBUNTUKVM: '775:root:syslog',
    NODE_TYPE_HOST_RHELKVM: '755:root:root',
    NODE_TYPE_EDGE: '775:root:syslog',
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY: '775:root:syslog',
    NODE_TYPE_UA: '775:root:syslog',
    NODE_TYPE_GLOBAL_MANAGER: '775:root:syslog',
}

NODE_TYPE_LOG_FILE_STATS = {
    NODE_TYPE_HOST_UBUNTUKVM: '640:syslog:adm',
    NODE_TYPE_HOST_RHELKVM: '640:root:root',
    NODE_TYPE_EDGE: '640:syslog:adm',
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY: '640:syslog:adm',
    NODE_TYPE_UA: '640:syslog:adm',
    NODE_TYPE_GLOBAL_MANAGER: '640:syslog:adm',
}

ROOT = 'root'
NEW_LINE_PATTERN = "\n"
MULTIPLE_SPACES_PATTERN = r"\s{2,}"
USR_SBIN_RSYSLOGD = '/usr/sbin/rsyslogd'
GET_RSYSLOG_USER_CMD_STR = 'ps -eo user,cmd | grep rsyslog'

STAT_CMD = ['stat', '-c', '%a:%U:%G']
if _NODE_TYPE == NODE_TYPE_HOST_ESX:
    _ROTATE_RSYSLOG = ['esxcli', 'system', 'syslog', 'reload']
elif _NODE_TYPE == NODE_TYPE_HOST_RHELKVM:
    _ROTATE_RSYSLOG = ["sudo", "service", "rsyslog", "force-reload"]
else:
    _ROTATE_RSYSLOG = ["sudo", "service", "rsyslog", "rotate"]
WRITE_VALID_STATE = ['7', '6', '3', '2']
EXECUTE_VALID_STATE = ['7', '5', '3', '1']
# this constant value cannot be a valid posix username(because posix
# usernames cannot have spaces). So use it to represent we can't get
# username from GET_RSYSLOG_USER_CMD
NO_USER = 'no user'


class AuditLogHealthForLocalErrorStatus(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._error_status = False

    def get_rsyslog_user(self):
        try:
            output = run_command(GET_RSYSLOG_USER_CMD_STR, shell=True) or ''
            self._logger.debug('get_rsyslog_user output: {}'.format(output))
            output_arr = output.split(NEW_LINE_PATTERN)
            for cur_line in output_arr:
                cur_res = re.split(MULTIPLE_SPACES_PATTERN, cur_line)
                if len(cur_res) == 2 and \
                        cur_res[1].startswith(USR_SBIN_RSYSLOGD):
                    self._logger.debug('rsyslog user: {}'.format(cur_res[0]))
                    return cur_res[0]
        except Exception as e:
            self._logger.warning("Failed to run command: {}, error: {}"
                                 .format(e, e.stderr))
            return None
        return NO_USER

    def get_stat(self, log_file_path):
        try:
            stat = run_command(STAT_CMD + [log_file_path],
                               shell=False) or ''
        except Exception as e:
            self._logger.warning("Failed to run command: {}, error: {}"
                                 .format(e, e.stderr))
            return None
        return stat

    def is_dir_in_invalid_state(self, cur_stat, target_stat):
        self._logger.debug("cur_stat: {}, target_stat: {}".format(
            cur_stat, target_stat))
        # state format is permission:owner:group
        cur_stat_arr = cur_stat.split(':')
        target_stat_arr = target_stat.split(':')
        # get the permission and add the omitted zero back
        cur_permission = list(cur_stat_arr[0])
        while len(cur_permission) < 3:
            cur_permission.insert(0, '0')
        self._logger.debug("cur_permission: {}".format(cur_permission))
        # consider 4 digit permission
        if cur_permission[-2] not in EXECUTE_VALID_STATE:
            return True
        return False

    def is_file_in_invalid_state(self, cur_stat, target_stat):
        self._logger.debug("cur_stat: {}, target_stat: {}".format(
            cur_stat, target_stat))
        # state format is permission:owner:group
        cur_stat_arr = cur_stat.split(':')
        target_stat_arr = target_stat.split(':')
        # get the permission and add the omitted zero back
        cur_permission = list(cur_stat_arr[0])
        while len(cur_permission) < 3:
            cur_permission.insert(0, '0')
        self._logger.debug("cur_permission: {}".format(cur_permission))
        # consider 4 digit permission
        if cur_permission[-3] not in WRITE_VALID_STATE:
            return True
        if cur_stat_arr[1] != target_stat_arr[1] \
            or cur_stat_arr[2] != target_stat_arr[2]:
            return True
        return False

    def rotate_rsyslog(self):
        self._logger.debug("rotate rsyslog")
        try:
            run_command(_ROTATE_RSYSLOG, shell=False)
        except Exception as e:
            self._logger.warning("rsyslog fails to rotate. Error: {}".format(
                e.stderr))
            return False
        return True

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        self._logger.debug("node_type: {}".format(_NODE_TYPE))
        # don't need to check if _NODE_TYPE is not in the _SUPPORTED_NODE_TYPES
        if _NODE_TYPE not in _SUPPORTED_NODE_TYPES:
            self._error_status = False
            return self._error_status
        # user cannot change log files' permission and ownership in esx
        # so only do the rsyslog roatate if any checked log file is missing
        # then return false
        if _NODE_TYPE == NODE_TYPE_HOST_ESX:
            for cur_log_file in NODE_TYPE_LOGS[_NODE_TYPE]:
                # check if cur_log_file exist
                cur_log_file_path = os.path.join(VAR_LOG_DIR, cur_log_file)
                if not os.path.exists(cur_log_file_path):
                    # Fix for Bug 2587505
                    # we will not raise alarm if the log is missing now,
                    # we will trigger a rsyslog rotate here instead to prevent
                    # the log is deleted by the user accidently
                    # "rotate rsyslog" part will be removed after a cron job
                    # (do rotate rsyslog) is added in the logging component
                    self.rotate_rsyslog()
                    break
            self._error_status = False
            return self._error_status
        cur_rsyslog_user = self.get_rsyslog_user()
        # cannot get rsyslog's user means rsyslog is not running,
        # so return error_status with true
        if cur_rsyslog_user is NO_USER:
            self._error_status = True
            return self._error_status
        # if get rsyslog user cmd failed, not update _error_status
        # stop check this time
        elif cur_rsyslog_user is None:
            return self._error_status

        # rsyslog's user is root, logs can always be written to the files
        # so only do the rsyslog roatate if any checked log file is missing
        # then return false
        if cur_rsyslog_user == ROOT:
            for cur_log_file in NODE_TYPE_LOGS[_NODE_TYPE]:
                # check if cur_log_file exist
                cur_log_file_path = os.path.join(VAR_LOG_DIR, cur_log_file)
                if not os.path.exists(cur_log_file_path):
                    # Fix for Bug 2587505
                    # we will not raise alarm if the log is missing now,
                    # we will trigger a rsyslog rotate here instead to prevent
                    # the log is deleted by the user accidently
                    # "rotate rsyslog" part will be removed after a cron job
                    # (do rotate rsyslog) is added in the logging component
                    self.rotate_rsyslog()
                    break
            self._error_status = False
            return self._error_status

        log_dir_stat = self.get_stat(VAR_LOG_DIR).strip()
        if log_dir_stat:
            if self.is_dir_in_invalid_state(log_dir_stat,
                            NODE_TYPE_LOG_DIR_STATS[_NODE_TYPE]):
                # if status changed, reload the rsyslog config
                if not self._error_status:
                    rotate_rsyslog_res = self.rotate_rsyslog()
                    # if rsyslog rotate failed, not update _error_status
                    # stop check this time
                    if not rotate_rsyslog_res:
                        return self._error_status
                self._logger.debug("{} permission and ownership is {}, "
                    "it should be {}".format(VAR_LOG_DIR, log_dir_stat,
                    NODE_TYPE_LOG_DIR_STATS[_NODE_TYPE]))
                self._error_status = True
                return self._error_status
        for cur_log_file in NODE_TYPE_LOGS[_NODE_TYPE]:
            # check if cur_log_file exist
            cur_log_file_path = os.path.join(VAR_LOG_DIR, cur_log_file)
            if not os.path.exists(cur_log_file_path):
                # Fix for Bug 2587505, 2691293
                # We will raise alarm only if rsyslog.log is missing. For
                # other missing logs, we will trigger a rsyslog rotate
                # here to prevent the log is deleted by the user accidently.
                # "rotate rsyslog" part will be removed after a cron job
                # (do rotate rsyslog) is added in the logging component
                if cur_log_file is RSYSLOG_LOG:
                    self._error_status = True
                    return self._error_status
                else:
                    rotate_rsyslog_res = self.rotate_rsyslog()
                # if rsyslog rotate failed, not update _error_status
                # stop check this time
                if not rotate_rsyslog_res:
                    return self._error_status
                continue
            cur_log_file_stat = self.get_stat(cur_log_file_path).strip()
            # if can not get this log file's status, just skip and continue
            # to check other log files
            if cur_log_file_stat:
                # check if cur_log_file's permission and ownership is correct
                # if not, return True
                if self.is_file_in_invalid_state(cur_log_file_stat,
                            NODE_TYPE_LOG_FILE_STATS[_NODE_TYPE]):
                    # if status changed, reload the rsyslog config
                    if not self._error_status:
                        rotate_rsyslog_res = self.rotate_rsyslog()
                        # if rsyslog rotate failed, not update _error_status,
                        # continue to check other files
                        if not rotate_rsyslog_res:
                            continue
                    self._logger.debug("{} permission and ownership is {}, "
                        "it should be {}".format(
                            cur_log_file_path, cur_log_file_stat,
                            NODE_TYPE_LOG_FILE_STATS[_NODE_TYPE]))
                    self._error_status = True
                    return self._error_status
        # if status changed, reload the rsyslog config
        if self._error_status:
            rotate_rsyslog_res = self.rotate_rsyslog()
            # if rsyslog rotate failed, not update _error_status
            if not rotate_rsyslog_res:
                return self._error_status
        self._error_status = False
        return self._error_status
