# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import sha.config
import traceback
import xml.etree.ElementTree as ET
import uuid
from nsx_utils.node_utils import get_node_type
from nsx_utils.constants import NODE_TYPE_HOST_ESX
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache


if sha.config.SYSTEM == sha.config.PLATFORM_WINDOWS:
    _APPLIANCE_INFO_XML = os.path.join(sha.config.NSX_DIR, "appliance-info.xml")
else:
    _APPLIANCE_INFO_XML = "/etc/vmware/nsx/appliance-info.xml"

_NODE_TYPE = get_node_type()
if _NODE_TYPE == NODE_TYPE_HOST_ESX:
    _LOOKUP_CMD = ["/bin/nslookup"]
elif os.path.exists("/usr/bin/nslookup"):
    _LOOKUP_CMD = ["sudo", "/usr/bin/nslookup"]
else:
    # if nslookup is not installed by default, use getenv hosts instead
    _LOOKUP_CMD = ["sudo", "/usr/bin/getent", "hosts"]


class ManagerFqdnLookupFailureStatus(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._status_dict = {}

    def _get_manager_info(self):
        manager_info = []
        if os.path.isfile(_APPLIANCE_INFO_XML):
            tree = ET.parse(_APPLIANCE_INFO_XML)
            for proxy in tree.iter(tag = "appliance-proxy"):
                fqdn = proxy.find("fqdn")
                manager_info.append({
                    "ip": proxy.find("ip").text,
                    "fqdn": fqdn.text if fqdn is not None else "",
                    "uuid": proxy.find("uuid").text
                })
        return manager_info

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        manager_info = self._get_manager_info()
        self._logger.debug("ManagerFqdnLookupFailureStatus: manager_info = %s",
            str(manager_info))
        status_dict = {}
        for info in manager_info:
            node_uuid = uuid.UUID(info["uuid"])
            node_fqdn = info["fqdn"] or ""
            context = {
                "entity_id": info["uuid"],
                "appliance_fqdn": node_fqdn
            }
            status_dict[node_uuid] = (False, context)
            if node_fqdn and node_fqdn != info["ip"]:
                try:
                    out = run_command(_LOOKUP_CMD + [node_fqdn]) or ''
                    self._logger.debug("ManagerFqdnLookupFailureStatus: "
                        "lookup output for %s is %s", str(node_fqdn), str(out))
                except Exception as e:
                    self._logger.warning("ManagerFqdnLookupFailureStatus: "
                        "DNS lookup on manager FQDN {} failed with error "
                        "reason: %s", node_fqdn, traceback.format_exc())
                    out = None
                # Set status to True if lookup failed or if the IP address
                # returned does not contain the IP address found in
                # appliance-info.xml
                if not out or info["ip"] not in out:
                    status_dict[node_uuid] = (True, context)
        # Delete old manager UUIDs in cases where node UUIDs changed
        deleted_entity_list =\
            list(set(self._status_dict.keys()) - set(status_dict.keys()))
        # Emit a log when there is a status change or entity id change
        if status_dict != self._status_dict or deleted_entity_list:
            self._logger.info("ManagerFqdnLookupFailureStatus: "
                "status_dict = %s, deleted_entity_list = %s",
                str(status_dict), str(deleted_entity_list))
        self._status_dict = status_dict
        return self._status_dict, deleted_entity_list
