# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import platform
import socket
import traceback
import sha.config
from threading import Thread
from sha.contrib.monitor.metricsReceiver.internal.message import Parser,ClientStopException
from sha.contrib.monitor.metricsReceiver.internal.client import Client,ClientChild

class ConnectionThread(Thread):
    """
    Connection thread to receive client's data.
    """
    def __init__(self, connection, client_address, queue):
        Thread.__init__(self)
        self._logger = sha.config.get_logger(__name__)
        self._address = client_address
        self._conn = connection
        self._stop = False
        self._client = ClientChild()
        self._parser = Parser(self._conn, self._client)
        self._queue = queue

    def run(self):
        while not self._stop:
            try:
                self._parser.recv()
            except ClientStopException as e:
                self._logger.warning("MM:The client is stopped")
                self._close(socket.SHUT_RDWR)
                break
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("MM:The connection has been closed by client{}:{}".format(e, trace))
                if self._client:
                    self._client.stop()
                self._close(socket.SHUT_RD)
                break
        self._logger.info("MM:ConnectionThread quit")

    def _close(self, how):
        self._logger.info("MM:Close connection {}".format(self._conn))
        if self._conn:
            self._conn.shutdown(how)
            self._conn.close()
            self._conn = None
        self._logger.info("MM:Close connection done")

    def stop(self):
        self._logger.info("MM:Stop connection")
        self._stop = True
        self._close(socket.SHUT_RDWR)


class Connection:
    """
    Connection to receive clients' connection.
    """
    SOCKET_PATH = "/var/run/vmware/nsx-sha/sha_socket"
    SOCKET_ADDR = "127.0.0.1"
    SOCKET_PORT = 11111
    SOCKET_CONNECTION_NUM = 20

    _logger = sha.config.get_logger(__name__)
    _connections = []
    _sock = None
    _stop = True
    _queue = None

    @classmethod
    def start(cls, queue):
        cls._queue = queue
        if platform.system() == sha.config.PLATFORM_LINUX:
            import os
            import pwd
            import grp
            uid = pwd.getpwnam("nsx-sha").pw_uid
            gid = grp.getgrnam("nsx-sha").gr_gid
            # create socket file
            try:
                os.unlink(cls.SOCKET_PATH)
            except OSError:
                if os.path.exists(cls.SOCKET_PATH):
                    raise

            cls._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            cls._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            cls._sock.bind(cls.SOCKET_PATH)
            os.chmod(cls.SOCKET_PATH, 0o770)
            os.chown(cls.SOCKET_PATH, uid, gid)
        else:
            cls._sock = socket.socket(socket.AF_INET,  socket.SOCK_STREAM)
            cls._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            cls._sock.bind((cls.SOCKET_ADDR, cls.SOCKET_PORT))
        cls._logger.info("MM:Connection sock {}".format(cls._sock))
        cls._stop = False

    @classmethod
    def is_connected(cls):
        return cls._sock is not None

    @classmethod
    def wait_for_data(cls):
        if not cls._sock:
            return None
        cls._sock.listen(Connection.SOCKET_CONNECTION_NUM)
        while not cls._stop:
            try:
                conn, addr = cls._sock.accept()
                cls._logger.info("MM:new connection {}.{} is accepted".format(conn, addr))
                new_connection = ConnectionThread(conn, addr, cls._queue)
                new_connection.start()
                cls._connections.append(new_connection)
            except Exception as e:
                trace = traceback.format_exc()
                cls._logger.warning("MM:Failed to wait for new connection{}:{}".format(e, trace))

    @classmethod
    def close(cls):
        cls._logger.info("MM:Close the connections")
        cls._stop = True
        """Stop accepting new connections"""
        if cls._sock:
            cls._sock.shutdown(socket.SHUT_RD)
        """Stop the existed connections"""
        for connection in cls._connections:
            connection.stop()
        """Close the socket"""
        if cls._sock:
            cls._sock.close()
            cls._sock = None
            cls._logger.info("MM:Socket is closed")
