# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import sha.config
from sha.contrib.metric.utils import run_command

import traceback
from xml.dom.minidom import Document

FIREWALL_RULESET_FILE_PATH = "/etc/vmware/firewall/"
FIREWALL_RULESET_FILE_NAME = "-Ruleset.xml"
FIREWALL_REFRESH_CMD = ['localcli', 'network', 'firewall', 'refresh']
FIREWALL_ENABLE_CMD = ['localcli', 'network', 'firewall', 'ruleset', 'set', '-e', '1', '-r']
FIREWALL_DISABLE_CMD = ['localcli', 'network', 'firewall', 'ruleset', 'set', '-e', '0', '-r']
FIREWALL_LIST_CMD = ['localcli', 'network', 'firewall', 'ruleset', 'list', '-r']

def putChildWithText(doc, parent, element, text):
    child = doc.createElement(element)
    child_text = doc.createTextNode(text)
    child.appendChild(child_text)
    parent.appendChild(child)

class ESXFirewall:
    """Firewall rules.
       E.g:
        <ConfigRoot>
           <service id="0302">
              <id>nsx-sha</id>
              <rule id='0000'>
                 <direction>outbound</direction>
                 <protocol>tcp</protocol>
                 <porttype>dst</porttype>
                 <port>5002</port>
              </rule>
              <enabled>false</enabled>
              <required>false</required>
           </service>
        </ConfigRoot>
    """
    _logger = sha.config.get_logger(__name__)

    @classmethod
    def generate_ruleset(cls, name, id, portList):
        doc = Document()
        root = doc.createElement('ConfigRoot')
        doc.appendChild(root)
        service = doc.createElement('service')
        root.appendChild(service)
        service.setAttribute('id', id)
        putChildWithText(doc, service, 'id', name)
        putChildWithText(doc, service, 'enabled', 'false')
        putChildWithText(doc, service, 'required', 'false')

        index = 0
        for port in portList:
            if port is None:
                continue
            rule = doc.createElement('rule')
            service.appendChild(rule)
            rule.setAttribute('id', '%04d' % index)
            putChildWithText(doc, rule, 'direction', 'outbound')
            putChildWithText(doc, rule, 'protocol', port['protocol'])
            putChildWithText(doc, rule, 'porttype', 'dst')
            putChildWithText(doc, rule, 'port', '%d' % port['port'])
            index += 1

        xml_str = doc.toprettyxml(indent='   ')
        cls._logger.debug('Firewall config: {}'.format(xml_str))

        # write rule to xml file
        firewall_file = FIREWALL_RULESET_FILE_PATH + name + FIREWALL_RULESET_FILE_NAME
        with open(firewall_file, 'w') as f:
            f.write(xml_str)
        cls._logger.info('Write firewall ruleset into file {} :\n{}'.format(firewall_file, xml_str))

    @classmethod
    def update_ruleset(cls, ruleset_name):
        firewall_enable_cmd = list(FIREWALL_ENABLE_CMD)
        firewall_enable_cmd.append(ruleset_name)
        cls._logger.debug("FIREWALL_ENABLE_CMD:{}, firewall_enable_cmd :{}"
                          .format(FIREWALL_ENABLE_CMD, firewall_enable_cmd))
        run_command(FIREWALL_REFRESH_CMD, shell=False)
        run_command(firewall_enable_cmd, shell=False)
        cls._logger.info("Firewall ruleset {} updated".format(ruleset_name))

    @classmethod
    def clear_ruleset(cls, ruleset_name):
        firewall_disable_cmd = list(FIREWALL_DISABLE_CMD)
        firewall_disable_cmd.append(ruleset_name)
        cls._logger.debug("FIREWALL_DISABLE_CMD:{}, firewall_disable_cmd :{}"
                          .format(FIREWALL_DISABLE_CMD, firewall_disable_cmd))
        run_command(firewall_disable_cmd, shell=False)
        cls._logger.info("Firewall ruleset {} disabled".format(ruleset_name))

    @classmethod
    def is_ruleset_enabled(cls, ruleset_name):
        firewall_list_cmd = list(FIREWALL_LIST_CMD)
        firewall_list_cmd.append(ruleset_name)
        try:
            output = run_command(firewall_list_cmd, shell=False)
            cls._logger.info("Check ruleset {}: \n{}".format(ruleset_name, output))
        except Exception as e:
            cls._logger.info("ruleset {} doesn't exist".format(ruleset_name))
            return False

        # localcli network firewall ruleset list -r nsx-sha
        # Name     Enabled
        # ----------------
        # nsx-sha  false
        return bool(output.splitlines()[2].split()[1])


IPTABLES_CMD = ['sudo', 'iptables']
IPTABLES_ENABLE_ARG = ['-A']
IPTABLES_DISABLE_ARG = [ '-D']
IPTABLES_CHECK_ARG = ['-C']
IPTABLES_LIST_ARG = ['-S']
IPTABLES_ACTION = ['-j', 'ACCEPT']
IPTABLES_OWNER_AGRS = ['-m', 'owner', '--uid-owner']

class LinuxFirewall:
    """Linux Frewall rules.
        E.g:
          iptables -C OUTPUT -p udp --dport 6831 -j ACCEPT
          iptables -A OUTPUT -p udp --dport 6831 -j ACCEPT
          iptables -D OUTPUT -p udp --dport 6831 -j ACCEPT
    """
    _logger = sha.config.get_logger(__name__)

    @classmethod
    def is_rule_enabled(cls, port, owner=None):
        options = [
            'OUTPUT',
            '-p', port['protocol'],
            '--dport',  '%d' % port['port'],
        ]
        if owner:
            options += IPTABLES_OWNER_AGRS + [owner]
        iptable_check_cmd = IPTABLES_CMD + IPTABLES_CHECK_ARG + options + IPTABLES_ACTION
        try:
            run_command(iptable_check_cmd, shell=False)
            cls._logger.info("Firewall rule '{}' exists".format(options + IPTABLES_ACTION))
            return True
        except Exception as e:
            cls._logger.info("Firewall rule '{}' doesn't exist".format(options + IPTABLES_ACTION))
            return False

    @classmethod
    def add_rules(cls, portList, owner=None):
        for port in portList:
            if cls.is_rule_enabled(port, owner):
                continue

            options = [
                'OUTPUT',
                '-p', port['protocol'],
                '--dport',  '%d' % port['port'],
            ]
            if owner:
                options += IPTABLES_OWNER_AGRS + [owner]
            iptable_enable_cmd = IPTABLES_CMD + IPTABLES_ENABLE_ARG + options + IPTABLES_ACTION
            run_command(iptable_enable_cmd, shell=False)
            cls._logger.info("Firewall rule '{}' added".format(options + IPTABLES_ACTION))

    @classmethod
    def remove_rules(cls, portList, owner=None):
        for port in portList:
            options = [
                'OUTPUT',
                '-p', port['protocol'],
                '--dport',  '%d' % port['port'],
            ]
            if owner:
                options += IPTABLES_OWNER_AGRS + [owner]
            iptable_disable_cmd = IPTABLES_CMD + IPTABLES_DISABLE_ARG + options + IPTABLES_ACTION
            run_command(iptable_disable_cmd, shell=False)
            cls._logger.info("Firewall rule '{}' removed".format(options + IPTABLES_ACTION))


class SHAFirewall:
    """Firewall rules for SHA.
    """
    def __init__(self, feature, ruleset_id, rule_owner):
        self._logger = sha.config.get_logger(__name__)
        self._feature = feature
        self._firewall_ruleset_name = "nsx-sha-" + feature
        self._firewall_ruleset_id = ruleset_id
        self._firewall_rule_owner = rule_owner
        self._firewall_linux_rules = []
        self._firewall_enabled = False
        # clear legacy firewall rule on ESX
        if sha.config.NODE_TYPE == sha.config.DIST_ESX:
            if ESXFirewall.is_ruleset_enabled(self._firewall_ruleset_name):
                try:
                    ESXFirewall.clear_ruleset(self._firewall_ruleset_name)
                except Exception as e:
                    self._logger.warning('Failed to clear legacy firewall rule for wavefront: {}'.format(e))
        # For linux, it is hard to know the legacy firewall rules from which configuration.
        # When adding new firewall rule, we will check whether existed firstly.

    def update(self, rules):
        if sha.config.NODE_TYPE != sha.config.DIST_ESX and \
           sha.config.SYSTEM != sha.config.PLATFORM_LINUX:
            return

        if sha.config.NODE_TYPE == sha.config.DIST_ESX:
            # empty rules, clear firewall rule
            if not rules:
                if self._firewall_enabled:
                    try:
                        ESXFirewall.clear_ruleset(self._firewall_ruleset_name)
                        self._firewall_enabled = False
                    except Exception as e:
                        self._logger.warning('Failed to clear firewall rule for {}: {}'.format(self._feature, e))
                return

            # write firewall ruleset xml file
            ESXFirewall.generate_ruleset(self._firewall_ruleset_name, self._firewall_ruleset_id, rules)

            # update firewall rules
            try:
                ESXFirewall.update_ruleset(self._firewall_ruleset_name)
                self._firewall_enabled = True
            except Exception as e:
                self._logger.warning('Failed to update firewall rule for {}: {}'.format(self._feature, e))

        elif sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
            # Remove old rules firstly
            if self._firewall_enabled and self._firewall_linux_rules:
                try:
                    LinuxFirewall.remove_rules(self._firewall_linux_rules, self._firewall_rule_owner)
                    self._firewall_enabled = False
                    self._firewall_linux_rules = []
                except Exception as e:
                    self._logger.warning('Failed to clear firewall rule for {}: {}'.format(self._feature, e))

            # add new rules if need
            if rules:
                try:
                    LinuxFirewall.add_rules(rules, self._firewall_rule_owner)
                    self._firewall_linux_rules = rules
                    self._firewall_enabled = True
                except Exception as e:
                    self._logger.warning('Failed to add firewall rule for {}: {}'.format(self._feature, e))
