# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import os
import ipaddress
from sha.core.channel.provider.wf_provider import WavefrontSender
import sha.config
from sha.config import PRE_CONDITION
from sha.constants import CONDITION_WAVEFRONT, CONDITION_TRACE, REACHED, PRECONDITION_KEY, \
                          COLLECTOR_TYPE_WAVE_FRONT, COLLECTOR_IP, COLLECTOR_PORT, COLLECTOR_TRACING_PORT, \
                          COLLECTOR_METRICS_PORT, COLLECTOR_HOST


from sha.core.whiteboard import DataWhiteboard
from sha.core.channel.provider.firewall_ruleset import SHAFirewall

if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
    from sha.core.channel.provider.tsdb_provider import TsdbSender

class GlobalConfigResolver:
    """
       Resolve GlobalConfigMsg from nestdb
    """
    _logger = sha.config.get_logger(__name__)

    @classmethod
    def get(cls):
        """
        get the global config
        """
        return {'trace_type': TraceResolver.get()}

    @classmethod
    def set(cls, msg, old_msg):
        """
        :param msg: GlobalConfigMsg
        :param old_msg: GlobalConfigMsg
        """
        # Set collector config
        if msg.HasField('netopa_collector_config'):
            CollectorResolver.set(msg.netopa_collector_config)
        elif old_msg.HasField('netopa_collector_config'):
            CollectorResolver.clear()
        # Set workflow trace config
        if msg.HasField('workflow_trace_config'):
            TraceResolver.set(msg.workflow_trace_config)
        elif old_msg.HasField('workflow_trace_config'):
            TraceResolver.clear()
        # Set TSDB config
        if msg.HasField('tsdb_global_config'):
            TSDBResolver.set(msg.tsdb_global_config)
        elif old_msg.HasField('tsdb_global_config'):
            TSDBResolver.clear()

    @classmethod
    def clear(cls, old_msg):
        """
        :param old_msg: GlobalConfigMsg
        """
        # Clear collector config
        if old_msg.HasField('netopa_collector_config'):
            CollectorResolver.clear()
        # Clear workflow trace config
        if old_msg.HasField('workflow_trace_config'):
            TraceResolver.clear()
        # Clear TSDB config
        if old_msg.HasField('tsdb_global_config'):
            TSDBResolver.clear()


class TSDBResolver:
    _logger = sha.config.get_logger(__name__)
    _enabled = False

    @classmethod
    def set(cls, tsdb_config):
        """
        :param tsdb_config: TsdbGlobalConfigMsg from nestdb
                            or TsdbGlobalConfigMsg from UA
        """
        if tsdb_config.HasField('enabled'):
            if cls._enabled == tsdb_config.enabled:
                return
            cls._enabled = tsdb_config.enabled
            cls._logger.info("TSDB config is {}.".format("enabled" if cls._enabled else "disabled"))
            # Only update TSDB on supported platforms
            if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
                TsdbSender.update_conf({'enabled': cls._enabled})
            else:
                cls._logger.info("TSDB not supported yet")
        else:
            cls._enabled = False
            cls._logger.info("TSDB config is cleared.")
            # Only update TSDB on supported platforms
            if sha.config.SYSTEM == sha.config.PLATFORM_LINUX:
                TsdbSender.update_conf()

    @classmethod
    def clear(cls):
        TsdbSender.update_conf()


WORKFLOW_TRACE_PROTO = 'udp'
WORKFLOW_TRACE_PORT = 6831
WORFFLOW_TRACE_RULE = {
                          'protocol':WORKFLOW_TRACE_PROTO,
                          'port': WORKFLOW_TRACE_PORT
                      }

class TraceResolver:
    _logger = sha.config.get_logger(__name__)
    _trace_type = 0
    _enabled = False
    if sha.config.is_ua() or sha.config.is_gm():
        # TODO: MP-CCP schema is going away with UFO, migrate away
        #       from this to nestdb_schema_common_pb2 which is packaged
        #       separately as nestdb-schema-common-py-proto-pkg
        from vmware.nsx.config import ops_config_pb2 as schema_common_pb2
    else:
        from vmware.nsx.nestdb import nestdb_schema_common_pb2 as schema_common_pb2
    _mapper = {
        schema_common_pb2.WorkFlowTraceConfigMsg.TraceType.Value('TRACE_TYPE_INVALID'): 0,
        schema_common_pb2.WorkFlowTraceConfigMsg.TraceType.Value('LOCAL'): 1,
        schema_common_pb2.WorkFlowTraceConfigMsg.TraceType.Value('JAEGER'): 2,
        schema_common_pb2.WorkFlowTraceConfigMsg.TraceType.Value('WAVE_FRONT'): 3,
    }
    _trace_firewall = SHAFirewall("trace", "0302", None)


    @classmethod
    def get(cls):
        """
        get the trace type
        """
        return cls._trace_type

    @classmethod
    def _update_pre_condition(cls):
        if PRE_CONDITION[CONDITION_TRACE][REACHED] != cls._enabled:
            PRE_CONDITION[CONDITION_TRACE][REACHED] = cls._enabled
            cls._logger.info("sha precondition changed: {}".format(PRE_CONDITION))
            if not DataWhiteboard.report_data(PRECONDITION_KEY, PRE_CONDITION):
                cls._logger.info('failed to report {} to whiteboard'.format(PRECONDITION_KEY))


    @classmethod
    def _set(cls, trace_config):
        """
        :param trace_config: WorkFlowTraceConfigMsg
        """
        if trace_config.HasField('trace_type'):
            cls._trace_type = cls._mapper.get(trace_config.trace_type)
        else:
            # no trace_type, clear it
            cls._trace_type = 0

        if trace_config.HasField('enabled'):
            cls._enabled = trace_config.enabled
        else:
            # no enabled, clear it
            cls._enabled = False

    @classmethod
    def _clear(cls):
        """
        :param old_trace_config: WorkFlowTraceConfigMsg
        """
        cls._trace_type = 0
        cls._enabled = False

    @classmethod
    def set(cls, trace_config):
        """
        :param trace_config: WorkFlowTraceConfigMsg
        """
        cls._set(trace_config)
        cls._logger.info("Workflow trace is set: trace_type({}), enabled({})"
                         .format(cls._trace_type, cls._enabled))
        # Update pre_conditioin
        cls._update_pre_condition()
        # Update firewall
        cls._trace_firewall.update([WORFFLOW_TRACE_RULE] if cls._trace_type == 2 and cls._enabled else [])

    @classmethod
    def clear(cls):
        """
        :param old_trace_config: WorkFlowTraceConfigMsg
        """
        cls._clear()
        cls._logger.info("Workflow trace is cleared: trace_type({}), enabled({})"
                         .format(cls._trace_type, cls._enabled))
        # Update pre_conditioin
        cls._update_pre_condition()
        # Update firewall
        cls._trace_firewall.update([])

class CollectorResolver:
    _logger = sha.config.get_logger(__name__)
    _collectors = dict()
    if sha.config.is_ua() or sha.config.is_gm():
        # TODO: MP-CCP schema is going away with UFO, migrate away
        #       from this to nestdb_schema_common_pb2 which is packaged
        #       separately as nestdb-schema-common-py-proto-pkg
        from vmware.nsx.config import ops_config_pb2 as schema_common_pb2
    else:
        from vmware.nsx.nestdb import nestdb_schema_common_pb2 as schema_common_pb2
    _mapper = {
        schema_common_pb2.NetopaCollectorMsg.CollectorType.Value(COLLECTOR_TYPE_WAVE_FRONT): COLLECTOR_TYPE_WAVE_FRONT,
    }

    @classmethod
    def get(cls, type):
        """
        get the collector of the specified type
        """
        return cls._collectors[type]

    @classmethod
    def _set(cls, sha_collector_config):
        """
        Add or update collectors, remove the legacy ones.
        :param sha_collector_config: NetopaCollectorConfigMsg
        """
        # Only one collector allowed for each type, record types of configured collectors
        configured_types = []
        for item in sha_collector_config.collector:
            cls._logger.debug("item in collector msg: {}".format(item))
            if not item.HasField('type'):
                continue
            type = cls._mapper.get(item.type, None)
            if type == COLLECTOR_TYPE_WAVE_FRONT:
                collector = cls._collectors.setdefault(type, Collector(type))
                if item.HasField(COLLECTOR_IP):
                    collector.set(COLLECTOR_IP, item.collector_ip)
                if item.HasField(COLLECTOR_PORT):
                    collector.set(COLLECTOR_PORT, item.collector_port)
                if item.HasField(COLLECTOR_TRACING_PORT):
                    collector.set(COLLECTOR_TRACING_PORT, item.tracing_port)
                collector.update()
                cls._logger.debug("wavefront collector:{}".format(collector))
                configured_types.append(type)

        # Remove legacy collectors, excepts the new configured ones
        cls._clear(configured_types)

    @classmethod
    def _clear(cls, active_collectors=[]):
        """
        Clear all collectors expects ones in active collector list
        :param active_collectors: a list of active collector types.
        """
        for type in list(cls._collectors.keys()):
            if type in active_collectors:
                 continue
            collector = cls._collectors[type]
            cls._logger.info("clear {} collector: {}.".format(type, collector.get()))
            collector.clear()
            del cls._collectors[type]


    @classmethod
    def set(cls, collector_msg):
        """
        :param collector_msg: NetopaCollectorConfigMsg
        """
        # add/update collectors, remove legacy ones
        cls._set(collector_msg)
        cls._logger.info('Netopa collectors:{}'.format(cls._collectors))

    @classmethod
    def clear(cls):
        """
        Clear all exsiting collectors
        """
        cls._clear()


class Collector:
    """The collector config.
    """
    def __init__(self, type):
        self._logger = sha.config.get_logger(__name__)
        self._type = type
        self._attrs = dict()
        self._changed = False

    def set(self, attr, val):
        if (attr not in self._attrs) or \
           (self._attrs[attr] != val):
            self._changed = True
            self._attrs[attr] = val

    def get(self):
        return self._attrs

    def _update_pre_condition(self, condition, reached):
        if PRE_CONDITION[condition][REACHED] != reached:
            PRE_CONDITION[condition][REACHED] = reached
            self._logger.info("sha precondition changed: {}".format(PRE_CONDITION))
            if not DataWhiteboard.report_data(PRECONDITION_KEY, PRE_CONDITION):
                self._logger.info('failed to report {} to whiteboard'.format(PRECONDITION_KEY))

    def update(self):
        if self._changed is not True:
            return
        if self._type == COLLECTOR_TYPE_WAVE_FRONT:
            try:
                # Update wavefront conf
                self._logger.info('Update wavefront collector: {}'.format(self._attrs))
                wavefront_conf = {}
                if COLLECTOR_IP in self._attrs:
                    if self._attrs[COLLECTOR_IP].ipv4 is not None:
                        wavefront_conf[COLLECTOR_HOST] = str(ipaddress.IPv4Address(self._attrs[COLLECTOR_IP].ipv4))
                    elif self._attrs[COLLECTOR_IP].ipv6 is not None:
                        wavefront_conf[COLLECTOR_HOST] = str(ipaddress.IPv6Address(self._attrs[COLLECTOR_IP].ipv6))
                if COLLECTOR_PORT in self._attrs:
                    wavefront_conf[COLLECTOR_METRICS_PORT] = int(self._attrs[COLLECTOR_PORT])
                if COLLECTOR_TRACING_PORT in self._attrs:
                    wavefront_conf[COLLECTOR_TRACING_PORT] = int(self._attrs[COLLECTOR_TRACING_PORT])
                self._logger.info("Wavefront conf: {}".format(wavefront_conf))
                # Update wavefront sender
                WavefrontSender.update(wavefront_conf)
                # Update pre_conditioin
                self._update_pre_condition(CONDITION_WAVEFRONT, True)
            except Exception as e :
                import traceback; traceback.print_exc(); print(e)
        self._changed = False

    def clear(self):
        self._attrs.clear()
        if self._type == COLLECTOR_TYPE_WAVE_FRONT:
            WavefrontSender.clear()
            self._update_pre_condition(CONDITION_WAVEFRONT, False)

